<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2020 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.3/2027974
 */

namespace TestIntegration\Service;

use Api\IRequest;
use Application\Config\ConfigManager;
use Application\Config\IDao;
use Application\Log\SwarmLogger;
use Application\View\Helper\ViewHelperFactory;
use Application\Escaper\Escaper;
use Interop\Container\ContainerInterface;
use P4\Model\Fielded\Iterator;
use Projects\Model\Project as ProjectModel;
use Record\Key\AbstractKey;
use TestIntegration\Controller\TestRunApi;
use TestIntegration\Filter\ITestRun;
use TestIntegration\Filter\StatusValidator;
use TestIntegration\Model\ITestDefinition;
use TestIntegration\Model\ITestRun as ITestRunModel;
use TestIntegration\Model\TestRun;
use Zend\Http\Client as HttpClient;
use Zend\Http\Request;
use Zend\Http\Response;

/**
 * Class TestExecutor
 * @package TestIntegration\Service
 */
class TestExecutor implements ITestExecutor
{
    private $services;
    private $logger;
    private $testRunDAO;
    private $review;
    private $currentProjects;
    private $projects;

    // The minimum API version that will support the new callback urls
    const API_VERSION = 'v10';

    use ProjectTestNameTrait;

    /**
     * TestExecutor constructor.
     * @inheritDoc
     */
    public function __construct(ContainerInterface $services, array $options = null)
    {
        $this->services   = $services;
        $this->logger     = $services->get(SwarmLogger::SERVICE);
        $this->testRunDAO = $services->get(IDao::TEST_RUN_DAO);
    }

    /**
     * @inheritDoc
     */
    public function runReviewTests(AbstractKey $review, array $currentProjects, Iterator $projects)
    {
        $this->review          = $review;
        $this->currentProjects = $currentProjects;
        $this->projects        = $projects;
        $this->runGlobalTests();
        $this->runProjectTests();
    }

    /**
     * @inheritDoc
     */
    public function runGlobalTests()
    {
        $review = $this->review;
        // Get the enabled global test definitions
        $testDefinitionDAO = $this->services->get(IDao::TEST_DEFINITION_DAO);
        $testDefinitions   = $testDefinitionDAO->fetchAll();

        // Set review vars that will not change within the loop
        $change       = $review->getHeadChange(true);
        $status       = $review->isPending() ? 'shelved' : 'submitted';
        $reviewId     = $review->getId();
        $version      = $review->getHeadVersion();
        $reviewToken  = $review->getToken();
        $reviewStatus = $review->getState();
        $description  = $this->sanitize($review->getDescription());
        $projectInfo  = $this->getProjectInfo();


        // Run a test for each global test definition
        foreach ($testDefinitions as $testDefinition) {
            // Create the test run record
            $testRun = $this->createTestRun($testDefinition->getName());

            // Test run could not be created, so we continue on
            if (is_null($testRun)) {
                continue;
            }

            // Get the callback urls
            $testRunId    = $testRun->getId();
            $callbackUrls = $this->generateCallbackUrls($testRunId, $reviewToken);

            // Used to build out the CI request url
            $fields = [
                '{change}'       => $change,
                '{status}'       => $status,
                '{review}'       => $reviewId,
                '{version}'      => $version,
                '{test}'         => $testRun->getTest(),
                '{testRunId}'    => $testRunId,
                '{projects}'     => $projectInfo['projectIds'],
                '{projectNames}' => $projectInfo['projectNames'],
                '{branches}'     => $projectInfo['branchIds'],
                '{pass}'         => $callbackUrls['pass'],
                '{fail}'         => $callbackUrls['fail'],
                '{update}'       => $callbackUrls['update'],
                '{reviewStatus}' => $reviewStatus,
                '{description}'  => $description
            ];

            $this->runGlobalTest($testDefinition, $fields, $testRun);
        }
    }

    /**
     * @inheritDoc
     */
    public function runProjectTests()
    {
        $review = $this->review;

        // compose deploy success/fail callback urls.
        $urlHelper        = $this->services->get('ViewHelperManager')->get(ViewHelperFactory::QUALIFIED_URL);
        $deploySuccessUrl = $urlHelper(
            'review-deploy',
            ['review' => $review->getId(), 'token' => $review->getToken(), 'status' => 'success']
        );
        $deployFailUrl    = $urlHelper(
            'review-deploy',
            ['review' => $review->getId(), 'token' => $review->getToken(), 'status' => 'fail']
        );

        // Set review vars that will not change within the loop
        $change         = $review->getHeadChange(true);
        $status         = $review->isPending() ? 'shelved' : 'submitted';
        $reviewId       = $review->getId();
        $version        = $review->getHeadVersion();
        $reviewToken    = $review->getToken();
        $description    = $this->sanitize($review->getDescription());
        $reviewStatus   = $review->getState();
        $testStartTimes = [];

        // reviews can impact multiple projects and each project can have its own test config
        // note we only include projects/branches the change being processed impacts.
        foreach ($this->currentProjects as $projectId => $branchIds) {
            // get the full project object and the list of impacted branch names
            $project     = isset($this->projects[$projectId]) ? $this->projects[$projectId] : null;
            $branchNames = $this->getBranchNamesFromIds($project, $branchIds);

            // if enabled and configured, run automated tests
            if ($project && $project->getTests('enabled') && $project->getTests('url')) {
                $testRun = $this->createTestRun($this->getProjectTestName($projectId));

                // Test run could not be created, so we continue on
                if (is_null($testRun)) {
                    continue;
                }

                // Get the callback urls
                $testRunId    = $testRun->getId();
                $callbackUrls = $this->generateCallbackUrls($testRunId, $reviewToken);

                $fields = [
                    '{change}'        => $change,
                    '{status}'        => $status,
                    '{review}'        => $reviewId,
                    '{version}'       => $version,
                    '{test}'          => $testRun->getTest(),
                    '{testRunId}'     => $testRunId,
                    '{project}'       => $project->getId(),
                    '{projectName}'   => $project->getName(),
                    '{branch}'        => implode(',', $branchIds),
                    '{branchName}'    => implode(',', $branchNames),
                    '{pass}'          => $callbackUrls['pass'],
                    '{fail}'          => $callbackUrls['fail'],
                    '{update}'        => $callbackUrls['update'],
                    '{deploySuccess}' => $deploySuccessUrl,
                    '{deployFail}'    => $deployFailUrl,
                    '{reviewStatus}'  => $reviewStatus,
                    '{description}'   => $description
                ];

                 $this->runProjectTest($project, $fields, $testRun, $testStartTimes);
            }

            // if enabled and configured, run deploy
            if ($project && $project->getDeploy('enabled') && $project->getDeploy('url')) {
                // customize test url for this project.
                $search  = [
                    '{change}',
                    '{status}',
                    '{review}',
                    '{project}',
                    '{projectName}',
                    '{branch}',
                    '{branchName}',
                    '{success}',
                    '{fail}'
                ];
                $replace = array_map(
                    'rawurlencode',
                    [
                        $change,
                        $status,
                        $reviewId,
                        $project->getId(),
                        $project->getName(),
                        implode(',', $branchIds),
                        implode(',', $branchNames),
                        $deploySuccessUrl,
                        $deployFailUrl
                    ]
                );

                $url = str_replace($search, $replace, $project->getDeploy('url'));
                $this->doRequest($url, 'automated deploy', ProjectModel::FORMAT_URL);
            }
        }

        $details = $review->getTestDetails(true);
        if ($testStartTimes || count($details['startTimes']) > count($details['endTimes'])) {
            $review->setTestDetails(['startTimes' => $testStartTimes, 'endTimes' => []] + $details)->save();
        }
    }

    /**
     * @inheritDoc
     */
    public function runGlobalTest(ITestDefinition $testDefinition, array $fields, ITestRunModel $testRun)
    {
        // We never allow the description in the url, so we preemptively remove it
        $url = str_replace('{description}', '', $testDefinition->getUrl());

        // Build out the necessary parameters for making the CI request
        $search   = array_keys($fields);
        $replace  = array_map('rawurlencode', array_values($fields));
        $url      = str_replace($search, $replace, $url);
        $postBody = str_replace($search, $replace, trim($testDefinition->getBody()));
        $headers  = $testDefinition->getHeaders() ?? [];
        $this->logger->info('Tests triggered for review: ' . $this->review->getId() . ' URL: ' . $url);
        $response = $this->doRequest(
            $url,
            $testDefinition->getTitle(),
            $testDefinition->getEncoding(),
            $postBody,
            $headers,
            $testDefinition->getTimeout()
        );
        $this->validateResponse($testRun, $url, $response);
    }

    /**
     * @inheritDoc
     */
    public function runProjectTest(ProjectModel $project, array $fields, ITestRunModel $testRun, array &$testStartTimes)
    {
        $review = $this->review;
        // We never allow the description in the url, so we preemptively remove it
        $url = str_replace('{description}', '', $project->getTests('url'));

        // Build out the necessary parameters for making the CI request
        $search      = array_keys($fields);
        $replace     = array_map('rawurlencode', array_values($fields));
        $url         = str_replace($search, $replace, $url);
        $postBody    = str_replace($search, $replace, trim($project->getTests('postBody')));
        $description = "Tests for " . $project->getId();
        $postFormat  = $project->getTests('postFormat') ?? ProjectModel::FORMAT_URL;
        $this->logger->info('Tests triggered for review: ' . $review->getId() . ' URL: ' . $url);
        $response = $this->doRequest($url, $description, $postFormat, $postBody);

        if ($response && $response->isSuccess()) {
            $testStartTimes[] = time();
            $review->set('testStatus', StatusValidator::STATUS_RUNNING);
            $review->save();
        }

        $this->validateResponse($testRun, $url, $response);
    }

    /**
     * Processes the currentProjects and projects arrays to return lists of project ids & names and branch ids (which
     * are composed from both the project id and the branch id)
     * @return array
     */
    protected function getProjectInfo()
    {
        $projectIds   = [];
        $projectNames = [];
        $branchIds    = [];

        foreach ($this->currentProjects as $projectId => $currentBranchIds) {
            if (isset($this->projects[$projectId])) {
                $project        = $this->projects[$projectId];
                $projectIds[]   = $projectId;
                $projectNames[] = $project->getName();

                foreach ($currentBranchIds as $branchId) {
                    // Construct the branch id in the form of <projectId>:<branchId>
                    $branchIds[] = sprintf('%s:%s', $projectId, $branchId);
                }
            }
        }
        return [
            'projectIds'   => implode(',', $projectIds),
            'projectNames' => implode(',', $projectNames),
            'branchIds'    => implode(',', $branchIds)
        ];
    }

    /**
     * Convenience method to get a list of branch names, given branch ids
     *
     * @param ProjectModel $project
     * @param array        $branchIds
     *
     * @return array
     */
    protected function getBranchNamesFromIds(ProjectModel $project, array $branchIds)
    {
        $branches    = $project->getBranches();
        $branchNames = [];
        foreach ($branchIds as $branchId) {
            foreach ($branches as $branch) {
                if ($branch['id'] == $branchId) {
                    $branchNames[] = $branch['name'];
                    break;
                }
            }
        }

        return $branchNames;
    }

    /**
     * Creates a new TestRun instance
     * @param string      $testName   name to identify which test this is
     * @return mixed
     */
    protected function createTestRun(string $testName)
    {
        $review  = $this->review;
        $testRun = null;
        $data    = [
            TestRun::FIELD_CHANGE     => $review->getId(),
            TestRun::FIELD_VERSION    => $review->getHeadVersion(),
            TestRun::FIELD_TEST       => $testName,
            TestRun::FIELD_START_TIME => time(),
            TestRun::FIELD_STATUS     => StatusValidator::STATUS_RUNNING,
            TestRun::FIELD_UUID       => $review->getToken(),
        ];

        $filter = $this->services->get(ITestRun::NAME);
        $filter->setData($data);

        if ($filter->isValid()) {
            $testRun = new TestRun();
            $testRun = $testRun->set($data);
            $this->testRunDAO->save($testRun);
            $review->addTestRun($testRun->getId())->save();
        } else {
            $errors = $filter->getMessages();
            $this->logValidationErrors($errors);
        }

        return $testRun;
    }

    /**
     * Executes the actual CI request
     *
     * @param string   $url          url for the CI system to run the test
     * @param string   $description  description of the test
     * @param string   $postBody     request parameters that will be sent in the body or in the url
     * @param string   $postFormat   either url or json
     * @param array    $headers      headers, if specified
     * @param int      $timeout      optional timeout, defaults to 0 and is ignored if 0
     *
     * @return mixed
     */
    protected function doRequest(
        string $url,
        string $description,
        string $postFormat,
        string $postBody = '',
        array $headers = [],
        int $timeout = 0
    ) {
        $response = null;
        $logger   = $this->logger;

        // extract the http client options; including any special overrides for our host
        $options = $this->services->get(ConfigManager::CONFIG) + ['http_client_options' => []];
        $options = (array)$options['http_client_options'];

        // below function call is to encode the url path
        $url = $this->encodeUrlPath($url);
        
        // attempt a request for the given url to trigger tests.
        try {
            $client = new HttpClient;
            $client->setUri($url);
            $client->setHeaders($headers);

            // if we have post data, ensure we make a POST request
            if (!empty($postBody)) {
                $client->setMethod(Request::METHOD_POST);

                // parse body based on its format (URL or JSON)
                if (strcasecmp($postFormat, ProjectModel::FORMAT_URL) === 0) {
                    parse_str($postBody, $postParams);
                    $client->setParameterPost($postParams);
                } elseif (strcasecmp($postFormat, ProjectModel::FORMAT_JSON) === 0) {
                    $client->setEncType('application/json');
                    $client->setRawBody($postBody);
                }
            }

            // calculate options, including host based overrides, and set them
            if (isset($options['hosts'][$client->getUri()->getHost()])) {
                $options = (array)$options['hosts'][$client->getUri()->getHost()] + $options;
            }
            unset($options['hosts']);
            if ($timeout > 0) {
                $options['timeout'] = $timeout;
            }
            $client->setOptions($options);
            $logger->trace(get_class($this) . ': test dispatch request ' . var_export($client->getRequest(), true));
            // attempt trigger remote build.
            $response = $client->dispatch($client->getRequest());
        } catch (\Exception $e) {
            $logger->err($e);
        }
        return $response;
    }

    /**
     * Generate callback urls which can be used by the CI system to update a given test run
     *
     * @param int      $testRunId   id of the test run
     * @param string   $uuid        uuid of the test run
     *
     * @return array
     */
    protected function generateCallbackUrls($testRunId, $uuid)
    {
        $urlHelper = $this->services->get('ViewHelperManager')->get(ViewHelperFactory::QUALIFIED_URL);
        $params    = [
            IRequest::VERSION   => self::API_VERSION,
            TestRun::FIELD_ID   => $testRunId,
            TestRun::FIELD_UUID => $uuid
        ];

        return [
            'pass'   => $urlHelper(TestRunAPI::PASS_URL_ROUTE, $params),
            'fail'   => $urlHelper(TestRunAPI::FAIL_URL_ROUTE, $params),
            'update' => $urlHelper(TestRunAPI::UPDATE_NOAUTH_URL_ROUTE, $params)
        ];
    }

    /**
     * Handle the logging of filter validation errors
     * @param array $errors
     */
    protected function logValidationErrors(array $errors)
    {
        $message = "TestRun could not be created!\n";

        foreach ($errors as $field => $fieldErrors) {
            $message .= "$field => ";
            foreach ($fieldErrors as $type => $error) {
                $message .= "$type: $error\n";
            }
        }

        $this->logger->err($message);
    }

    /**
     * Ensure that the response from the CI system is valid
     * If the response is not valid we:
     *  - log that there was no response and if there was a response, add the reason and code
     *  - set the TestRun record's status as a fail
     *  - set the TestRun record's messages as 1-element array, containing a string indicating the code & reason for
     *    the validation failure or a message saying there was no response
     *
     * @param ITestRunModel   $testRun    TestRun record
     * @param string          $url        url of the CI system
     * @param Response|null   $response   response from the CI system or null
     */
    protected function validateResponse(ITestRunModel $testRun, string $url, $response)
    {
        // All good, nothing to do here, so exit early
        if ($response && $response->isSuccess()) {
            return;
        }
        if (is_null($response)) {
            $message = sprintf("There was no response from %s", explode('?', $url)[0]);
        } else {
            $message = sprintf("%s:%s", $response->getStatusCode(), $response->getReasonPhrase());
        }
        $this->logger->err(sprintf("Failed to trigger %s: %s, %s", $testRun->getTest(), $url, $message));
        $testRun->setStatus(StatusValidator::STATUS_FAIL);
        $testRun->setMessages([$message]);
        $this->testRunDAO->save($testRun);
    }

    /**
     * Escapes a string to deal with markdown, newlines and unicode
     * @param string   $dirty   string to be cleaned
     * @return string
     */
    protected function sanitize(string $dirty)
    {

        if (empty($dirty)) {
            $clean = '';
        } else {
            $parsedown = new \Parsedown();
            $markdown  = $parsedown->text($dirty);
            $stripped  = strip_tags($markdown);
            $clean     = htmlentities($stripped);
        }
        return $clean;
    }

    /**
     * It encodes the url path and return the new url
     * @param string   $url url for the CI system to run the test
     * @return string
     */
    protected function encodeUrlPath(string $url)
    {
        $escaper = new Escaper();
        $url     = $escaper->escapeFullUrl(rawurldecode($url));
        return $url;
    }
}
